import requests
import click
import sys

FIREBASE_API_KEY = "AIzaSyDcJB_1Or2TbRxTXajQCr-MTipwwQSiY38"

def refresh_id_token(refresh_token):
    data = {
        "grant_type": "refresh_token",
        "refresh_token": refresh_token,
    }

    try:
        response = requests.post(
            f"https://securetoken.googleapis.com/v1/token?key={FIREBASE_API_KEY}",
            data=data,
            timeout=5,
        )

        # Attempt to decode the response to JSON before calling raise_for_status
        # to ensure we can inspect the error response body if needed.
        response_json = response.json()

        # If the response status indicates success, check for the id_token.
        if response.ok and response_json.get("id_token"):
            return (
                response_json["id_token"],
                response_json["refresh_token"],
                response_json["user_id"],
            )

        # If the response is not OK, but we've successfully parsed the JSON,
        # we can now check for known error conditions without raising an exception.
        error_message = response_json.get("error", {}).get("message")
        known_errors = [
            "TOKEN_EXPIRED",
            "USER_DISABLED",
            "USER_NOT_FOUND",
            "INVALID_REFRESH_TOKEN",
            "INVALID_GRANT_TYPE",
            "MISSING_REFRESH_TOKEN",
        ]

        if error_message in known_errors:
            return None, None, None

        # If the error is not one of the known errors, then raise an HTTPError
        response.raise_for_status()

    except requests.exceptions.HTTPError as e:
        # If we catch an HTTPError, we check if it's due to one of the known errors after raise_for_status() call
        if response.status_code == 400:
            return None, None, None
        else:
            # For unexpected HTTP errors, log the error.
            click.echo(f"HTTP request error: {e}")
            raise e
    except requests.exceptions.RequestException as e:
        # For other types of requests exceptions, log the error.
        click.echo(f"Request error: {e}")
        raise e
    except ValueError:
        # If the response couldn't be decoded as JSON, log the error.
        click.echo("Invalid JSON response.")
        raise

    return None, None, None


def is_token_valid(id_token):
    data = {"idToken": id_token}

    try:
        response = requests.post(
            f"https://identitytoolkit.googleapis.com/v1/accounts:lookup?key={FIREBASE_API_KEY}",
            json=data,
            timeout=5,
        )

        # If the request was successful, the token is valid.
        if response.status_code == 200:
            return True

        # Handle potential JSON response even in error cases
        response_json = response.json()
        error_message = response_json.get("error", {}).get("message")

        # Check for specific error messages indicating invalid token or user not found
        if error_message in ("INVALID_ID_TOKEN", "USER_NOT_FOUND"):
            return None

    except requests.exceptions.RequestException as e:
        click.echo(f"Request error: {e}")
    except ValueError:
        click.echo("Invalid JSON response.")

    # Return False for other errors not explicitly handled above
    return False


import click
from thunder import auth
import sys
import os
from thunder import thunder_helper
from thunder import container_helper
from importlib.metadata import version
import requests
from packaging import version as version_parser
from thunder import api

PACKAGE_NAME = "tnr"  # update if name changes
# Get the directory of the current file (thunder.py), then go up two levels to the root
current_dir = os.path.dirname(os.path.abspath(__file__))
root_dir = os.path.join(current_dir, "..", "..")

# Add the src directory to sys.path
sys.path.append(root_dir)


class DefaultCommandGroup(click.Group):
    def resolve_command(self, ctx, args: list):
        try:
            # Try to resolve the command normally
            check_for_update()
            return super(DefaultCommandGroup, self).resolve_command(ctx, args)
        except click.exceptions.UsageError:
            # If no command is found, default to 'run' and include the args
            return "run", run, args


@click.group(
    cls=DefaultCommandGroup,
    help="This CLI is the interface between you and Thunder Compute.",
    invoke_without_command=True,
)
@click.option('--version', '--v', is_flag=True, help="Show the CLI version.")
@click.pass_context
def cli(ctx, version):
    if version:
        click.echo(f"{PACKAGE_NAME} version {get_version()}")
        ctx.exit()
    if ctx.invoked_subcommand is None:
        click.echo(ctx.get_help())
        ctx.exit()


@cli.command(
    help="Runs a specified task on Thunder Compute. This is the default behavior of the tnr command. Please see thundergpu.net for detailed documentation.",
)
@click.option("--ngpus", default=1, help="Specify the number of GPUs to use.")
@click.argument("args", nargs=-1)
def run(ngpus, args):
    """
    Executes a task on Thunder Compute.

    Args:
        ngpus: The number of GPUs to use for the task.
        args: Additional arguments to pass to the task.
    """
    if not args:  # Check if args is empty
        click.echo("No arguments provided. Exiting...")
        sys.exit(0)

    id_token, refresh_token, uid = auth.load_tokens()

    if not id_token or not refresh_token:
        click.echo("Please log in to begin using Thunder Compute.")
        id_token, refresh_token, uid = auth.login()

        if not id_token or not refresh_token:
            return
    
    if not api.is_token_valid(id_token):
        id_token, refresh_token, uid = auth.handle_token_refresh(refresh_token)


    # Create an instance of Task with required arguments
    task = thunder_helper.Task(ngpus, args, uid)

    # Initialize the session with the current auth token
    if not task.get_password(id_token):
        id_token, refresh_token, uid = auth.handle_token_refresh(refresh_token)
        if not id_token or not refresh_token or not uid:
            click.echo("Token, Refresh Token, or UID is invalid. Please log in again.")
        if not task.get_password(id_token):
            click.echo("Failed to retrieve password for session.")
            return

    # Execute the task
    if not task.execute_task(id_token):
        return

    # Close the session
    if not task.close_session(id_token):
        click.echo("Failed to close the session.")

@click.option('--port', '-p', multiple=True, help="Port(s) to forward from the container. To specify a continuous range use <start_port>:<end_port> syntax. To specify multiple port / ranges add another --port / -p flag.")
@cli.command(
    help="Creates a Docker container for running commands on Thunder Compute in MacOS and Windows. If you are having trouble in your linux environment, this command may help."
)
def container(port):
    """
    Creates a Docker container for running commands on Thunder Compute in MacOS and Windows. If you are having trouble in your linux environment, this command may help.
    """
    container_helper.create_docker_container(port)


@cli.command(help="Logs in to Thunder Compute.")
def login():
    """
    Authenticates a user for access to Thunder Compute. You can use this to create a new account or to log in to an existing account. Your login will also work for the Thunder Compute console (console.thundercompute.com).
    """
    auth.login()


@cli.command(help="Logs out from the Thunder Compute CLI.")
def logout():
    """
    Logs the current user out of the Thunder Compute CLI.
    """
    auth.logout()


def check_for_update():
    try:
        current_version = version(PACKAGE_NAME)
        response = requests.get(f"https://pypi.org/pypi/{PACKAGE_NAME}/json", timeout=1)
        json_data = response.json() if response else {}
        latest_version = json_data.get("info", {}).get("version", None)
        if version_parser.parse(current_version) < version_parser.parse(latest_version):
            click.echo(
                f"Update available: {latest_version}. Run `pip install --upgrade {PACKAGE_NAME}` to update.",
                err=True,
            )
    except Exception as e:
        click.echo(f"Error checking for update: {e}", err=True)


def get_version():
    return version(PACKAGE_NAME)


if __name__ == "__main__":
    # Custom argument handling before Click processing
    argv = sys.argv[1:]
    known_commands = ["login", "logout", "help", "run", "container"]

    if argv:
        if argv[0] not in known_commands:
            command_and_args = ["run", " ".join(argv)]
            argv = command_and_args
        if argv[0] == "run" and len(argv) > 1:
            command_and_args = ["run", " ".join(argv[1:])]
            argv = command_and_args

    # Now, manually invoke the CLI with the modified arguments
    sys.argv[1:] = argv
    cli()



import os
import sys
import click
import requests

sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))
from thunder.get_latest import get_latest

# URLs for cloud functions
CREATE_SESSION_USER_URL = "https://create-session-user-b7ngwrgpka-uc.a.run.app"
DELETE_SESSION_USER_URL = "https://delete-session-user-b7ngwrgpka-uc.a.run.app"


# Helper to call cloud functions
def call_firebase_function(url, id_token, payload=None):
    try:
        headers = {
            "Authorization": "Bearer " + id_token,
            "Content-Type": "application/json",
        }

        response = requests.post(url, headers=headers, json=payload, timeout=5)
        return response
    except Exception as _:
        msg = "Failed to create a session on the thunder network. Please report this issue to the developers!"
        click.echo(click.style(msg, fg='white', bg='red'))
        exit(1)


class Task:
    def __init__(self, ngpus: int, args: tuple, uid: str):
        self.ngpus = ngpus
        self.args = args[0].split(" ")
        self.username = uid
        self.password = None
        self.topic = None

    # set environment variable for
    # use firebase credentials to retrieve session password
    def get_password(self, id_token: str) -> bool:
        payload = {}
        response = call_firebase_function(CREATE_SESSION_USER_URL, id_token, payload)
        if response.status_code != 200:
            click.echo("Failed to create user.")
            return False
        response = response.json()
        if not isinstance(response, dict):
            click.echo("Failed to create user.")
            return False

        password = response.get("password")

        if not password:
            click.echo("Invalid response: Password missing.")
            return False

        self.password = password
        return True

    def execute_task(self, id_token: str) -> bool:
        # Get from https://storage.googleapis.com/thunder_binary/
        binary = get_latest("client", "~/.thunder/libthunder.so")
        if binary == None:
            print("Failed to download binary")
            return False

        os.environ["SESSION_USERNAME"] = self.username
        os.environ["SESSION_PASSWORD"] = self.password
        os.environ["TOKEN"] = id_token
        os.environ["LD_PRELOAD"] = f'{binary}'

        # This should never return
        try:
            os.execvp(self.args[0], self.args)
        except FileNotFoundError:
            click.echo(click.style(f"Failed to run the following command: {' '.join(self.args)}", fg='red', bold=True))
        except Exception as e:
            print(e)
        return False

    def close_session(self, id_token: str) -> bool:
        # to implement, communicate to manager computation is done, triggering exit_user
        click.echo("finishing task")
        return True



from setuptools import setup, find_packages

setup(
    name="tnr",
    version="1.2.4",
    package_dir={"": "src"},  # Specify the root directory for packages
    packages=find_packages(where="src"),  # Tell setuptools to find packages under src
    include_package_data=True,  # Include other files specified in MANIFEST.in
    install_requires=[
        "Click>=8.0",  # Specify a minimum version if needed
        "requests>=2.2",  # Same here, adjust the version as per your compatibility requirements
        "cryptography>=40.0",  # Adjust based on the features you're using
        "colorama>=0.4.0",  # Add a version constraint to ensure compatibility
        "distro>=1.7.0",  # Specify according to the functionalities you need
        "docker-shell>=0.0.9",  
    ],
    entry_points={"console_scripts": ["tnr=thunder.thunder:cli"]},
)

# delete old dist folder first, and increment version number

# to build: python3 setup.py sdist bdist_wheel
# to distribute: twine upload dist/*
